/*
 * Written by Doug Lea with assistance from members of JCP JSR-166
 * Expert Group and released to the public domain, as explained at
 * http://creativecommons.org/licenses/publicdomain
 */


/**
 * Preview versions of classes targetted for Java 7.
 */
package jsr166y;
